###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order membrane bioreactor model
"""
import pytest
from io import StringIO

from pyomo.environ import (
    Block, ConcreteModel, Constraint, value, Var, assert_optimal_termination)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import MBRZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestMBRZOdefault:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(default={"solute_list": ["tss",
                                                                    "nonvolatile_toc",
                                                                    "toc",
                                                                    "eeq",
                                                                    "viruses_enteric",
                                                                    "total_coliforms_fecal_ecoli",
                                                                    "cryptosporidium"]})

        m.fs.unit = MBRZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1000)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "nonvolatile_toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "eeq"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "cryptosporidium"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("mbr")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(1.007, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(0.99305, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.99305, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(0.99305, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["toc"]))
        assert (pytest.approx(0.99305, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["eeq"]))
        assert (pytest.approx(0.99305, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["viruses_enteric"]))
        assert (pytest.approx(0.99305, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["total_coliforms_fecal_ecoli"]))
        assert (pytest.approx(0.99305, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["cryptosporidium"]))
        assert (pytest.approx(1.001222762, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.49939, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.39951, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(0.29110405, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["toc"]))
        assert (pytest.approx(0.11985345, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["eeq"]))
        assert (pytest.approx(0.0099878, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["viruses_enteric"]))
        assert (pytest.approx(0.001200532, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["total_coliforms_fecal_ecoli"]))
        assert (pytest.approx(9.9878e-05, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["cryptosporidium"]))
        assert (pytest.approx(0.0057772, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(86.547, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]))
        assert (pytest.approx(103.8558564, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(122.6433808, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["toc"]))
        assert (pytest.approx(152.3219227, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["eeq"]))
        assert (pytest.approx(171.36216303, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["viruses_enteric"]))
        assert (pytest.approx(172.8850361, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["total_coliforms_fecal_ecoli"]))
        assert (pytest.approx(173.0757847, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["cryptosporidium"]))
        assert (pytest.approx(8580.775896, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                          : Value   : Fixed : Bounds
                              Electricity Demand :  8580.8 : False : (0, None)
                           Electricity Intensity :  2.3670 :  True : (None, None)
                Solute Removal [cryptosporidium] : 0.99990 :  True : (0, None)
                            Solute Removal [eeq] : 0.88000 :  True : (0, None)
                Solute Removal [nonvolatile_toc] : 0.60000 :  True : (0, None)
                            Solute Removal [toc] : 0.70854 :  True : (0, None)
    Solute Removal [total_coliforms_fecal_ecoli] : 0.99880 :  True : (0, None)
                            Solute Removal [tss] : 0.50000 :  True : (0, None)
                Solute Removal [viruses_enteric] : 0.99000 :  True : (0, None)
                                  Water Recovery : 0.99990 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                                     Inlet   Treated   Byproduct
    Volumetric Flowrate                             1.0070     1.0012 0.0057772 
    Mass Concentration H2O                          993.05     998.68    17.309 
    Mass Concentration tss                         0.99305    0.49939    86.547 
    Mass Concentration nonvolatile_toc             0.99305    0.39951    103.86 
    Mass Concentration toc                         0.99305    0.29110    122.64 
    Mass Concentration eeq                         0.99305    0.11985    152.32 
    Mass Concentration viruses_enteric             0.99305  0.0099878    171.36 
    Mass Concentration total_coliforms_fecal_ecoli 0.99305  0.0012005    172.89 
    Mass Concentration cryptosporidium             0.99305 9.9878e-05    173.08 
====================================================================================
"""

        assert output in stream.getvalue()

class TestMBRZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(default={"solute_list": ["tss",
                                                                    "nonvolatile_toc",
                                                                    "toc",
                                                                    "eeq",
                                                                    "viruses_enteric",
                                                                    "total_coliforms_fecal_ecoli",
                                                                    "cryptosporidium",
                                                                    "foo"]})

        m.fs.unit = MBRZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1000)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "nonvolatile_toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "eeq"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "cryptosporidium"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("mbr")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(1.008, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(0.99206, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.99206, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(0.99206, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["toc"]))
        assert (pytest.approx(0.99206, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["eeq"]))
        assert (pytest.approx(0.99206, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["viruses_enteric"]))
        assert (pytest.approx(0.99206, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["total_coliforms_fecal_ecoli"]))
        assert (pytest.approx(0.99206, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["cryptosporidium"]))
        assert (pytest.approx(0.99206, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["foo"]))
        assert (pytest.approx(1.002222762018698, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.49889, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(0.39911, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(0.290813, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["toc"]))
        assert (pytest.approx(0.119734, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["eeq"]))
        assert (pytest.approx(0.00997782, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["viruses_enteric"]))
        assert (pytest.approx(0.00119934, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["total_coliforms_fecal_ecoli"]))
        assert (pytest.approx(9.9788e-05, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["cryptosporidium"]))
        assert (pytest.approx(0.99778, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["foo"]))
        assert (pytest.approx(0.0057772, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(86.547, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["tss"]))
        assert (pytest.approx(103.8558564, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["nonvolatile_toc"]))
        assert (pytest.approx(122.6433808, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["toc"]))
        assert (pytest.approx(152.3219227, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["eeq"]))
        assert (pytest.approx(171.36216303, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["viruses_enteric"]))
        assert (pytest.approx(172.8850361, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["total_coliforms_fecal_ecoli"]))
        assert (pytest.approx(173.0757847, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["cryptosporidium"]))
        assert (pytest.approx(1.73093e-6, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]))
        assert (pytest.approx(8589.297024, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                          : Value   : Fixed : Bounds
                              Electricity Demand :  8589.3 : False : (0, None)
                           Electricity Intensity :  2.3670 :  True : (None, None)
                Solute Removal [cryptosporidium] : 0.99990 :  True : (0, None)
                            Solute Removal [eeq] : 0.88000 :  True : (0, None)
                            Solute Removal [foo] :  0.0000 :  True : (0, None)
                Solute Removal [nonvolatile_toc] : 0.60000 :  True : (0, None)
                            Solute Removal [toc] : 0.70854 :  True : (0, None)
    Solute Removal [total_coliforms_fecal_ecoli] : 0.99880 :  True : (0, None)
                            Solute Removal [tss] : 0.50000 :  True : (0, None)
                Solute Removal [viruses_enteric] : 0.99000 :  True : (0, None)
                                  Water Recovery : 0.99990 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                                     Inlet   Treated   Byproduct
    Volumetric Flowrate                             1.0080     1.0022  0.0057772
    Mass Concentration H2O                          992.06     997.68     17.309
    Mass Concentration tss                         0.99206    0.49889     86.547
    Mass Concentration nonvolatile_toc             0.99206    0.39911     103.86
    Mass Concentration toc                         0.99206    0.29081     122.64
    Mass Concentration eeq                         0.99206    0.11973     152.32
    Mass Concentration viruses_enteric             0.99206  0.0099778     171.36
    Mass Concentration total_coliforms_fecal_ecoli 0.99206  0.0011993     172.89
    Mass Concentration cryptosporidium             0.99206 9.9788e-05     173.08
    Mass Concentration foo                         0.99206    0.99778 1.7309e-06
====================================================================================
"""

        assert output in stream.getvalue()


db = Database()
params = db._get_technology("mbr")

class TestMBRZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(default={"solute_list": ["tss",
                                                                    "nonvolatile_toc",
                                                                    "toc",
                                                                    "eeq",
                                                                    "viruses_enteric",
                                                                    "total_coliforms_fecal_ecoli",
                                                                    "cryptosporidium"]})

        m.fs.unit = MBRZO(default={
            "property_package": m.fs.params,
            "database": db})

        return m

    @pytest.mark.parametrize("subtype", [k for k in params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("mbr", subtype=subtype)

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

@pytest.mark.parametrize("subtype", [k for k in params.keys()])
def test_costing(subtype):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tds"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = MBRZO(default={
        "property_package": m.fs.params,
        "database": m.db,
        "process_subtype": subtype})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tds"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.mbr, Block)
    assert isinstance(m.fs.costing.mbr.capital_a_parameter,Var)
    assert isinstance(m.fs.costing.mbr.capital_b_parameter,Var)
    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a granular activated carbon unit.
operation.
"""

from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_sido, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("GACZO")
class GACZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a granular activated carbon unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "gac"

        build_sido(self)

        # Empty Bed Contact Time
        self.empty_bed_contact_time = Var(
            units=pyunits.hour,
            bounds=(0, None),
            doc="Empty bed contact time of unit")

        self._fixed_perf_vars.append(self.empty_bed_contact_time)
        self._perf_var_dict["Empty Bed Contact Time"] = \
            self.empty_bed_contact_time

        # Electricity Demand
        self.electricity = Var(self.flowsheet().time,
                               units=pyunits.kW,
                               bounds=(0, None),
                               doc="Electricity consumption of unit")

        self.electricity_intensity_parameter = Var(
            units=pyunits.kW/pyunits.m**3,
            doc="Parameter for calculating electricity base on empty bed "
            "contacting time")

        self.energy_electric_flow_vol_inlet = Var(
            units=pyunits.kWh/pyunits.m**3,
            doc="Electricity intensity with respect to inlet flowrate of unit")

        @self.Constraint(
            doc='Electricity intensity based on empty bed contact time.')
        def electricity_intensity_constraint(b):
            return (b.energy_electric_flow_vol_inlet ==
                    b.electricity_intensity_parameter *
                    b.empty_bed_contact_time)

        @self.Constraint(self.flowsheet().time,
                         doc='Constraint for electricity consumption based on '
                         'feed flowrate.')
        def electricity_consumption(b, t):
            return b.electricity[t] == (
                b.energy_electric_flow_vol_inlet *
                pyunits.convert(b.get_inlet_flow(t),
                                to_units=pyunits.m**3/pyunits.hour))

        self._fixed_perf_vars.append(self.electricity_intensity_parameter)

        self._perf_var_dict["Electricity Demand"] = self.electricity
        self._perf_var_dict["Electricity Intensity"] = \
            self.energy_electric_flow_vol_inlet

        # Demand for activated carbon
        self.activated_carbon_replacement = Var(
            units=pyunits.kg/pyunits.m**3,
            bounds=(0, None),
            doc="Replacement rate of activated carbon")

        self.activated_carbon_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg/pyunits.hour,
            bounds=(0, None),
            doc="Demand for activated carbon")

        @self.Constraint(self.flowsheet().time,
                         doc='Constraint for activated carbon consumption.')
        def activated_carbon_equation(b, t):
            return b.activated_carbon_demand[t] == (
                b.activated_carbon_replacement *
                pyunits.convert(b.get_inlet_flow(t),
                                to_units=pyunits.m**3/pyunits.hour))

        self._fixed_perf_vars.append(self.activated_carbon_replacement)

        self._perf_var_dict["Activated Carbon Demand"] = \
            self.activated_carbon_demand



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order gas-sparged membrane unit
"""
import pytest
from io import StringIO

from idaes.core import declare_process_block_class, FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core.util import get_solver
import idaes.core.util.scaling as iscale
from idaes.config import bin_directory as idaes_bin_directory

from pyomo.environ import (check_optimal_termination,
                           ConcreteModel,
                           Constraint,
                           value,
                           Var)
from pyomo.network import Port
from pyomo.util.check_units import assert_units_consistent


from watertap.core import WaterParameterBlock, WaterStateBlock
from watertap.core.wt_database import Database
from watertap.unit_models.zero_order import GasSpargedMembraneZO
from watertap.unit_models.zero_order.gas_sparged_membrane_zo import initialize_sido, \
    calculate_scaling_factors_gas_extraction, _get_Q_gas_extraction

solver = get_solver()


class TestGasSpargedMembraneZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["cod"]})

        m.fs.unit = GasSpargedMembraneZO(
            default={"property_package": m.fs.water_props,
                     "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1000)
        m.fs.unit.inlet.flow_mass_comp[0, "cod"].fix(1)

        return m

    @pytest.mark.unit
    def test_private_attributes(self, model):
        assert model.fs.unit.config.database is model.db

        assert model.fs.unit._has_recovery_removal is True
        assert model.fs.unit._initialize is initialize_sido
        assert model.fs.unit._scaling is calculate_scaling_factors_gas_extraction
        assert model.fs.unit._get_Q is _get_Q_gas_extraction
        assert model.fs.unit._stream_table_dict == {
            "Inlet": model.fs.unit.inlet,
            "Treated": model.fs.unit.treated,
            "Byproduct": model.fs.unit.byproduct}
        assert model.fs.unit._perf_var_dict == {
            "Water Recovery": model.fs.unit.recovery_frac_mass_H2O,
            "Solute Removal": model.fs.unit.removal_frac_mass_solute,
            "Mass of gas extracted per mass flow of influent(kg/d/(kg/d)": model.fs.unit.gas_mass_influent_ratio,
            "Mass flow of gas extracted (kg/s))": model.fs.unit.flow_mass_gas_extraction,
            "Electricity Demand": model.fs.unit.electricity}

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.fs.unit.properties_in, WaterStateBlock)
        assert isinstance(model.fs.unit.properties_treated, WaterStateBlock)
        assert isinstance(model.fs.unit.properties_byproduct, WaterStateBlock)

        assert isinstance(model.fs.unit.inlet, Port)
        assert isinstance(model.fs.unit.treated, Port)
        assert isinstance(model.fs.unit.byproduct, Port)

        assert isinstance(model.fs.unit.gas_mass_influent_ratio, Var)
        assert len(model.fs.unit.gas_mass_influent_ratio) == 1
        assert isinstance(model.fs.unit.recovery_frac_mass_H2O, Var)
        assert len(model.fs.unit.recovery_frac_mass_H2O) == 1
        assert isinstance(model.fs.unit.removal_frac_mass_solute, Var)
        assert len(model.fs.unit.removal_frac_mass_solute) == 1

        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert len(model.fs.unit.water_recovery_equation) == 1
        assert isinstance(model.fs.unit.mass_balance, Constraint)
        assert len(model.fs.unit.mass_balance) == 1
        assert isinstance(model.fs.unit.solute_removal_equation, Constraint)
        assert len(model.fs.unit.solute_removal_equation) == 1
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)
        assert len(model.fs.unit.solute_treated_equation) == 1
        assert isinstance(model.fs.unit.mass_gas_extraction_equation, Constraint)
        assert len(model.fs.unit.mass_gas_extraction_equation) == 1

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("gas_sparged_membrane")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.gas_mass_influent_ratio[0].fixed
        assert model.fs.unit.gas_mass_influent_ratio[0].value == data[
            "gas_mass_influent_ratio"]["value"]

    @pytest.mark.unit
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model)

    @pytest.mark.component
    def test_scaling(self, model):
        iscale.calculate_scaling_factors(model)

        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.water_recovery_equation[0]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.mass_balance[0]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.solute_removal_equation[0, "cod"]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.solute_treated_equation[0, "cod"]) == 1e5

    @pytest.mark.component
    def test_initialization(self, model):
        initialization_tester(model)

    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(979, rel=1e-5) ==
                value(model.fs.unit.treated.flow_mass_comp[0, "H2O"]))
        assert (pytest.approx(20.58784, rel=1e-5) ==
                value(model.fs.unit.byproduct.flow_mass_comp[0, "H2O"]))
        assert (pytest.approx(0, abs=1e-5) ==
                value(model.fs.unit.treated.flow_mass_comp[0, "cod"]))
        assert (pytest.approx(1, rel=1e-5) ==
                value(model.fs.unit.byproduct.flow_mass_comp[0, "cod"]))

        assert (pytest.approx(1.001, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(0.9990001, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["cod"]))
        assert (pytest.approx(0.979, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0, abs=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["cod"]))
        assert (pytest.approx(0.021588, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(46.32238, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["cod"]))
        assert (pytest.approx(368.855328, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))
        assert (pytest.approx(5*value(model.fs.unit.properties_in[0].flow_vol)*0.08235, rel=1e-5) ==
                value(model.fs.unit.flow_mass_gas_extraction[0]))


    @pytest.mark.component
    def test_conservation(self, model):
        for (t, j) in model.fs.unit.inlet.flow_mass_comp.keys():
            if j != "H2O":
                assert (abs(value(model.fs.unit.inlet.flow_mass_comp[t, j] -
                                  model.fs.unit.treated.flow_mass_comp[t, j] -
                                  model.fs.unit.byproduct.flow_mass_comp[t, j]))
                        <= 1e-6)
            else:
                assert (abs(value(model.fs.unit.inlet.flow_mass_comp[t, j] -
                                  model.fs.unit.treated.flow_mass_comp[t, j] -
                                  model.fs.unit.byproduct.flow_mass_comp[t, j] -
                                  model.fs.unit.flow_mass_gas_extraction[t]))
                        <= 1e-6)

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()
        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                                         : Value      : Fixed : Bounds
                                             Electricity Demand :     368.86 : False : (0, None)
                             Mass flow of gas extracted (kg/s)) :    0.41216 : False : (0, None)
    Mass of gas extracted per mass flow of influent(kg/d/(kg/d) : 0.00041175 :  True : (0, None)
                                           Solute Removal [cod] :     1.0000 :  True : (0, None)
                                                 Water Recovery :    0.97900 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                             Inlet   Treated   Byproduct
    Volumetric Flowrate     1.0010    0.97900  0.021588 
    Mass Concentration H2O  999.00     1000.0    953.68 
    Mass Concentration cod 0.99900 1.0215e-15    46.322 
====================================================================================
"""

        assert output == stream.getvalue()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order VFA recovery model
"""
import pytest

from io import StringIO
from pyomo.environ import (
    ConcreteModel, Constraint, value, Var, assert_optimal_termination)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester

from watertap.unit_models.zero_order import VFARecoveryZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock

solver = get_solver()

class TestVFARecoveryZO_no_default:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["nonbiodegradable_cod"]})

        m.fs.unit = VFARecoveryZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "nonbiodegradable_cod"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("vfa_recovery")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(0.011, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(90.909090, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["nonbiodegradable_cod"]))
        assert (pytest.approx(0.0105, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(47.619047, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["nonbiodegradable_cod"]))
        assert (pytest.approx(0.0005, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(1000.0, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["nonbiodegradable_cod"]))
        assert (pytest.approx(0, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                   : Value      : Fixed : Bounds
                       Electricity Demand : 8.0000e-10 : False : (0, None)
                    Electricity Intensity :     0.0000 :  True : (None, None)
    Solute Removal [nonbiodegradable_cod] :    0.50000 :  True : (0, None)
                           Water Recovery :     1.0000 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                              Inlet   Treated  Byproduct
    Volumetric Flowrate                     0.011000 0.010500 0.00050000
    Mass Concentration H2O                    909.09   952.38 1.6000e-06
    Mass Concentration nonbiodegradable_cod   90.909   47.619     1000.0
====================================================================================
"""

        assert output in stream.getvalue()

class TestVFARecoveryZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["nonbiodegradable_cod", "foo"]})

        m.fs.unit = VFARecoveryZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "nonbiodegradable_cod"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("vfa_recovery")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(0.012, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(83.333, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["nonbiodegradable_cod"]))
        assert (pytest.approx(83.333, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["foo"]))
        assert (pytest.approx(0.0115, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(43.478261, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["nonbiodegradable_cod"]))
        assert (pytest.approx(86.95652, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["foo"]))
        assert (pytest.approx(0.0005, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(1000.0, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["nonbiodegradable_cod"]))
        assert (pytest.approx(1.6e-06, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["foo"]))
        assert (pytest.approx(0, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                   : Value      : Fixed : Bounds
                       Electricity Demand : 8.0000e-10 : False : (0, None)
                    Electricity Intensity :     0.0000 :  True : (None, None)
                     Solute Removal [foo] :     0.0000 :  True : (0, None)
    Solute Removal [nonbiodegradable_cod] :    0.50000 :  True : (0, None)
                           Water Recovery :     1.0000 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                              Inlet   Treated  Byproduct
    Volumetric Flowrate                     0.012000 0.011500 0.00050000
    Mass Concentration H2O                    833.33   869.57 1.6000e-06
    Mass Concentration nonbiodegradable_cod   83.333   43.478     1000.0
    Mass Concentration foo                    83.333   86.957 1.6000e-06
====================================================================================
"""

        assert output in stream.getvalue()

db = Database()
params = db._get_technology("vfa_recovery")


class Test_VFARecovery_ZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["nonbiodegradable_cod"]})

        m.fs.unit = VFARecoveryZO(default={
            "property_package": m.fs.params,
            "database": db})

        return m

    @pytest.mark.parametrize("subtype", [params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("vfa_recovery", subtype=subtype)

        model.fs.unit.load_parameters_from_database()

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order mabr model
"""
import pytest

from io import StringIO
from pyomo.environ import (
    Block, ConcreteModel, Constraint, value, Var, assert_optimal_termination, units as pyunits)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import MABRZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()

class TestMABRZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["bod",
                                     "tss",
                                     "ammonium_as_nitrogen",
                                     "nitrate"]})

        m.fs.unit = MABRZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "bod"].fix(5)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(5)
        m.fs.unit.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(2)
        m.fs.unit.inlet.flow_mass_comp[0, "nitrate"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("mabr")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j not in data['removal_frac_mass_solute'].keys():
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(0.023, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].flow_vol))
        assert (pytest.approx(217.3913, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["bod"]))
        assert (pytest.approx(217.3913, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["tss"]))
        assert (pytest.approx(86.9565, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["ammonium_as_nitrogen"]))
        assert (pytest.approx(43.4783, rel=1e-5) ==
                value(model.fs.unit.properties_in[0].conc_mass_comp["nitrate"]))

        assert (pytest.approx(0.023, rel=1e-2) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(217.3913, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["bod"]))
        assert (pytest.approx(104.3478, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].conc_mass_comp["nitrate"]))

        assert (pytest.approx(4e-12, rel=1e-2) ==
                value(model.fs.unit.properties_byproduct[0].flow_vol))
        assert (pytest.approx(200, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["bod"]))
        assert (pytest.approx(200, rel=1e-5) ==
                value(model.fs.unit.properties_byproduct[0].conc_mass_comp["nitrate"]))
        assert (pytest.approx(0.573804, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] +
                sum(model.fs.unit.generation_rxn_comp[0, r, j]
                    for r in model.fs.unit.reaction_set) -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                   : Value     : Fixed : Bounds
                        Blower Size (m^2) :    1.0000 :  True : (0, None)
                       Electricity Demand :   0.57380 : False : (0, None)
                    Electricity Intensity : 0.0069300 :  True : (None, None)
    Reaction Extent [ammonium_to_nitrate] :    1.4000 : False : (None, None)
    Solute Removal [ammonium_as_nitrogen] :    0.0000 :  True : (0, None)
                     Solute Removal [bod] :    0.0000 :  True : (0, None)
                 Solute Removal [nitrate] :    0.0000 :  True : (0, None)
                     Solute Removal [tss] :    0.0000 :  True : (0, None)
                           Water Recovery :    1.0000 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                              Inlet   Treated  Byproduct
    Volumetric Flowrate                     0.023000 0.023000 4.0000e-12
    Mass Concentration H2O                    434.78   434.78     200.00
    Mass Concentration bod                    217.39   217.39     200.00
    Mass Concentration tss                    217.39   217.39     200.00
    Mass Concentration ammonium_as_nitrogen   86.957   26.087     200.00
    Mass Concentration nitrate                43.478   104.35     200.00
====================================================================================
"""

        assert output in stream.getvalue()

def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["bod", "tss", "ammonium_as_nitrogen", "nitrate"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = MABRZO(default={
        "property_package": m.fs.params,
        "database": m.db})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10)
    m.fs.unit1.inlet.flow_mass_comp[0, "bod"].fix(5)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(5)
    m.fs.unit1.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "nitrate"].fix(1)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.mabr, Block)
    assert isinstance(m.fs.costing.mabr.specific_removal,
                      Var)
    assert isinstance(m.fs.costing.mabr.reactor_cost,
                      Var)
    assert isinstance(m.fs.costing.mabr.specific_air_flow,
                      Var)
    assert isinstance(m.fs.costing.mabr.blower_cost,
                      Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
General costing package for zero-order processes.
"""
import os
import yaml

import pyomo.environ as pyo
from pyomo.common.config import ConfigValue
from pyomo.util.calc_var_value import calculate_variable_from_constraint

from idaes.core import declare_process_block_class
from idaes.generic_models.costing.costing_base import (
    FlowsheetCostingBlockData, register_idaes_currency_units)

from watertap.core.zero_order_base import ZeroOrderBase
from watertap.unit_models.zero_order import (
    BrineConcentratorZO,
    ChemicalAdditionZO,
    ChlorinationZO,
    CoagulationFlocculationZO,
    DeepWellInjectionZO,
    DMBRZO,
    FixedBedZO,
    GACZO,
    LandfillZO,
    MABRZO,
    IonExchangeZO,
    IronManganeseRemovalZO,
    OzoneZO,
    OzoneAOPZO,
    SedimentationZO,
    StorageTankZO,
    SurfaceDischargeZO,
    UVZO,
    UVAOPZO,
    EvaporationPondZO,
    FilterPressZO,
    WellFieldZO,
    )


global_params = ["plant_lifetime",
                 "utilization_factor",
                 "land_cost_percent_FCI",
                 "working_capital_percent_FCI",
                 "salaries_percent_FCI",
                 "benefit_percent_of_salary",
                 "maintenance_costs_percent_FCI",
                 "laboratory_fees_percent_FCI",
                 "insurance_and_taxes_percent_FCI",
                 "wacc",
                 "TPEC",
                 "TIC"]


@declare_process_block_class("ZeroOrderCosting")
class ZeroOrderCostingData(FlowsheetCostingBlockData):

    CONFIG = FlowsheetCostingBlockData.CONFIG()
    CONFIG.declare("case_study_definition", ConfigValue(
        default=None,
        doc="Path to YAML file defining global parameters for case study. If "
        "not provided, default values from the WaterTap database are used."))

    def build_global_params(self):
        """
        To minimize overhead, only create global parameters for now.

        Unit-specific parameters will be added as sub-Blocks on a case-by-case
        basis as a unit of that type is costed.
        """
        # Load case study definition from file
        cs_def = _load_case_study_definition(self)

        # Register currency and conversion rates
        if "currency_definitions" in cs_def:
            pyo.units.load_definitions_from_strings(
                cs_def["currency_definitions"])
        else:
            register_idaes_currency_units()

        # Set the base year for all costs
        self.base_currency = getattr(pyo.units, cs_def["base_currency"])
        # Set a base period for all operating costs
        self.base_period = getattr(pyo.units, cs_def["base_period"])

        # Define expected flows
        self.defined_flows = {}
        for f, v in cs_def["defined_flows"].items():
            self.defined_flows[f] = v["value"]*getattr(pyo.units, v["units"])

        # Costing factors
        self.plant_lifetime = pyo.Var(units=self.base_period,
                                      doc="Plant lifetime")
        self.utilization_factor = pyo.Var(
            units=pyo.units.dimensionless,
            doc='Plant capacity utilization [%]')

        self.land_cost_percent_FCI = pyo.Var(units=pyo.units.dimensionless,
                                             doc="Land cost as % FCI")
        self.working_capital_percent_FCI = pyo.Var(
            units=pyo.units.dimensionless,
            doc="Working capital as % FCI")
        self.salaries_percent_FCI = pyo.Var(
            units=1/self.base_period,
            doc="Salaries as % FCI")
        self.benefit_percent_of_salary = pyo.Var(
            units=pyo.units.dimensionless,
            doc="Benefits as % salaries")
        self.maintenance_costs_percent_FCI = pyo.Var(
            units=1/self.base_period,
            doc="Maintenance and contingency costs as % FCI")
        self.laboratory_fees_percent_FCI = pyo.Var(
            units=1/self.base_period,
            doc="Laboratory fees as % FCI")
        self.insurance_and_taxes_percent_FCI = pyo.Var(
            units=1/self.base_period,
            doc="Insurance and taxes as % FCI")

        self.wacc = pyo.Var(units=pyo.units.dimensionless,
                            doc='Weighted Average Cost of Capital (WACC)')
        self.capital_recovery_factor = pyo.Expression(
            expr=((self.wacc *
                   (1 + self.wacc)**(self.plant_lifetime/self.base_period)) /
                  (((1 + self.wacc)**(self.plant_lifetime/self.base_period)) -
                   1) / self.base_period))

        self.TPEC = pyo.Var(units=pyo.units.dimensionless,
                            doc='Total Purchased Equipment Cost (TPEC)')
        self.TIC = pyo.Var(units=pyo.units.dimensionless,
                           doc='Total Installed Cost (TIC)')

        # Fix all Vars from database
        for v in global_params:
            try:
                value = cs_def["global_parameters"][v]["value"]
                units = cs_def["global_parameters"][v]["units"]
                getattr(self, v).fix(value*getattr(pyo.units, units))
            except KeyError:
                raise KeyError(
                    f"Invalid case study definition file - no entry found "
                    f"for {v}, or entry lacks value and units.")

    def build_process_costs(self):
        """
        Calculating process wide costs.
        """
        # Other capital costs
        self.land_cost = pyo.Var(
            initialize=0,
            units=self.base_currency,
            doc="Land costs - based on aggregate captial costs")
        self.working_capital = pyo.Var(
            initialize=0,
            units=self.base_currency,
            doc="Working capital - based on aggregate captial costs")
        self.total_capital_cost = pyo.Var(
            initialize=0,
            units=self.base_currency,
            doc="Total capital cost of process")

        self.land_cost_constraint = pyo.Constraint(
            expr=self.land_cost ==
            self.aggregate_capital_cost*self.land_cost_percent_FCI)
        self.working_capital_constraint = pyo.Constraint(
            expr=self.working_capital ==
            self.aggregate_capital_cost*self.working_capital_percent_FCI)
        self.total_capital_cost_constraint = pyo.Constraint(
            expr=self.total_capital_cost ==
            self.aggregate_capital_cost+self.land_cost+self.working_capital)

        # Other fixed costs
        self.salary_cost = pyo.Var(
            initialize=0,
            units=self.base_currency/self.base_period,
            doc="Salary costs - based on aggregate captial costs")
        self.benefits_cost = pyo.Var(
            initialize=0,
            units=self.base_currency/self.base_period,
            doc="Benefits costs - based on percentage of salary costs")
        self.maintenance_cost = pyo.Var(
            initialize=0,
            units=self.base_currency/self.base_period,
            doc="Maintenance costs - based on aggregate captial costs")
        self.laboratory_cost = pyo.Var(
            initialize=0,
            units=self.base_currency/self.base_period,
            doc="Laboratory costs - based on aggregate captial costs")
        self.insurance_and_taxes_cost = pyo.Var(
            initialize=0,
            units=self.base_currency/self.base_period,
            doc="Insurance and taxes costs - based on aggregate captial costs")
        self.total_fixed_operating_cost = pyo.Var(
            initialize=0,
            units=self.base_currency/self.base_period,
            doc="Total fixed operating costs")

        self.salary_cost_constraint = pyo.Constraint(
            expr=self.salary_cost ==
            self.aggregate_capital_cost*self.salaries_percent_FCI)
        self.benefits_cost_constraint = pyo.Constraint(
            expr=self.benefits_cost ==
            self.salary_cost*self.benefit_percent_of_salary)
        self.maintenance_cost_constraint = pyo.Constraint(
            expr=self.maintenance_cost ==
            self.aggregate_capital_cost*self.maintenance_costs_percent_FCI)
        self.laboratory_cost_constraint = pyo.Constraint(
            expr=self.laboratory_cost ==
            self.aggregate_capital_cost*self.laboratory_fees_percent_FCI)
        self.insurance_and_taxes_cost_constraint = pyo.Constraint(
            expr=self.insurance_and_taxes_cost ==
            self.aggregate_capital_cost*self.insurance_and_taxes_percent_FCI)

        self.total_fixed_operating_cost_constraint = pyo.Constraint(
            expr=self.total_fixed_operating_cost ==
            self.aggregate_fixed_operating_cost +
            self.salary_cost +
            self.benefits_cost +
            self.maintenance_cost +
            self.laboratory_cost +
            self.insurance_and_taxes_cost)

        # Other variable costs
        self.total_variable_operating_cost = pyo.Expression(
            expr=self.aggregate_variable_operating_cost +
            sum(self.aggregate_flow_costs[f] for f in self.flow_types),
            doc="Total variable operating cost of process per operating period")

        self.total_operating_cost = pyo.Expression(
            expr=(self.total_fixed_operating_cost +
                  self.total_variable_operating_cost),
            doc="Total operating cost of process per operating period")

    def initialize_build(self):
        """
        Basic initialization for flowsheet level quantities
        """
        calculate_variable_from_constraint(
            self.land_cost, self.land_cost_constraint)
        calculate_variable_from_constraint(
            self.working_capital, self.working_capital_constraint)
        calculate_variable_from_constraint(
            self.total_capital_cost, self.total_capital_cost_constraint)

        calculate_variable_from_constraint(
            self.salary_cost, self.salary_cost_constraint)
        calculate_variable_from_constraint(
            self.benefits_cost, self.benefits_cost_constraint)
        calculate_variable_from_constraint(
            self.maintenance_cost, self.maintenance_cost_constraint)
        calculate_variable_from_constraint(
            self.laboratory_cost, self.laboratory_cost_constraint)
        calculate_variable_from_constraint(
            self.insurance_and_taxes_cost,
            self.insurance_and_taxes_cost_constraint)

        calculate_variable_from_constraint(
            self.total_fixed_operating_cost,
            self.total_fixed_operating_cost_constraint)

    def add_LCOW(self, flow_rate):
        """
        Add Levelized Cost of Water (LCOW) to costing block.

        Args:
            flow_rate - flow rate of water (volumetric) to be used in
                        calculating LCOW
        """
        self.LCOW = pyo.Expression(
            expr=(self.total_capital_cost*self.capital_recovery_factor +
                  self.total_operating_cost) /
                 (pyo.units.convert(
                     flow_rate,
                     to_units=pyo.units.m**3/self.base_period) *
                  self.utilization_factor),
            doc='Levelized Cost of Water')

    def add_electricity_intensity(self, flow_rate):
        """
        Add calculation of overall electricity intensity to costing block.

        Args:
            flow_rate - flow rate of water (volumetric) to be used in
                        calculating electricity intensity
        """
        self.electricity_intensity = pyo.Expression(
            expr=pyo.units.convert(
                self.aggregate_flow_electricity / flow_rate,
                to_units=pyo.units.kWh/pyo.units.m**3),
            doc='Overall electricity intensity')

    # -------------------------------------------------------------------------
    # Unit operation costing methods
    def cost_power_law_flow(blk):
        """
        General method for costing equipment based on power law form. This is
        the most common costing form for zero-order models.

        CapCost = A*(F/Fref)**B

        This method also registers electricity demand as a costed flow (if
        present in the unit operation model).
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, state_ref = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter", "capital_b_parameter", "reference_state"])

        # Get state block for flow bases
        basis = parameter_dict["capital_cost"]["basis"]
        try:
            sblock = blk.unit_model.properties_in[t0]
        except AttributeError:
            # Pass-through case
            sblock = blk.unit_model.properties[t0]

        if basis == "flow_vol":
            state = sblock.flow_vol
            sizing_term = state/state_ref
        elif basis == "flow_mass":
            state = sum(sblock.flow_mass_comp[j]
                        for j in sblock.component_list)
            sizing_term = state/state_ref
        else:
            raise ValueError(
                f"{blk.name} - unrecognized basis in parameter "
                f"declaration: {basis}.")

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Call general power law costing method
        ZeroOrderCostingData._general_power_law_form(
            blk, A, B, sizing_term, factor)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_brine_concentrator(blk):
        """
        General method for costing brine concentration processes. Capital cost
        is based on the volumetirc flowrate and TDS of the incoming stream and
        the water recovery.

        This method also registers the electricity demand as a costed flow.
        """
        t0 = blk.flowsheet().time.first()
        inlet_state = blk.unit_model.properties_in[t0]

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C, D = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter",
             "capital_b_parameter",
             "capital_c_parameter",
             "capital_d_parameter"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        expr = (
            pyo.units.convert(
                A,
                to_units=blk.config.flowsheet_costing_block.base_currency) +
            pyo.units.convert(
                B*inlet_state.conc_mass_comp["tds"],
                to_units=blk.config.flowsheet_costing_block.base_currency) +
            pyo.units.convert(
                C*blk.unit_model.recovery_frac_mass_H2O[t0],
                to_units=blk.config.flowsheet_costing_block.base_currency) +
            pyo.units.convert(
                D*inlet_state.flow_vol,
                to_units=blk.config.flowsheet_costing_block.base_currency))

        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_chemical_addition(blk):
        """
        General method for costing chemical addition processes. Capital cost is
        based on the mass flow rate of chemical added.

        This method also registers the chemical flow and electricity demand as
        costed flows.
        """
        chem_name = blk.unit_model.config.process_subtype

        t0 = blk.flowsheet().time.first()
        chem_flow_mass = (blk.unit_model.chemical_dosage[t0] *
                          blk.unit_model.properties[t0].flow_vol /
                          blk.unit_model.ratio_in_solution)
        sizing_term = (blk.unit_model.chemical_flow_vol[t0] /
                       (pyo.units.gal/pyo.units.day))

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter", "capital_b_parameter"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Call general power law costing method
        ZeroOrderCostingData._general_power_law_form(
            blk, A, B, sizing_term, factor)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")
        blk.config.flowsheet_costing_block.cost_flow(
            chem_flow_mass, chem_name)

    def cost_chlorination(blk):
        """
        General method for costing chlorination units. Capital cost is based on
        the both inlet flow and dosage of chlorine.

        This method also registers the chemical flow and electricity demand as
        costed flows.
        """
        t0 = blk.flowsheet().time.first()
        chem_flow_mass = (blk.unit_model.chlorine_dose[t0] *
                          blk.unit_model.properties_in[t0].flow_vol)

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter",
             "capital_b_parameter",
             "capital_c_parameter"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        ln_Q = pyo.log(pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol /
            (pyo.units.m**3/pyo.units.hour),
            to_units=pyo.units.dimensionless))
        ln_D = pyo.log(pyo.units.convert(
            blk.unit_model.chlorine_dose[t0] / (pyo.units.mg/pyo.units.liter),
            to_units=pyo.units.dimensionless))

        expr = pyo.units.convert(
            A*ln_Q + B*ln_D + C*ln_Q*ln_D,
            to_units=blk.config.flowsheet_costing_block.base_currency)

        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")
        blk.config.flowsheet_costing_block.cost_flow(
            chem_flow_mass, "chlorine")

    def cost_coag_and_floc(blk):
        """
        General method for costing coagulation/flocculation processes. Capital cost
        is based on the alum flowrate and the polymer flowrate of the incoming stream.

        This method also registers the electricity demand as a costed flow.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C, D, E, F, G, H = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_mix_a_parameter",
             "capital_mix_b_parameter",
             "capital_floc_a_parameter",
             "capital_floc_b_parameter",
             "capital_coag_inj_a_parameter",
             "capital_coag_inj_b_parameter",
             "capital_floc_inj_a_parameter",
             "capital_floc_inj_b_parameter"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        cost_rapid_mix = (A*pyo.units.convert(
            blk.unit_model.rapid_mix_basin_vol,
            to_units=pyo.units.gallons) + B)*blk.unit_model.num_rapid_mix_processes

        cost_floc = (C*pyo.units.convert(
            blk.unit_model.floc_basin_vol,
            to_units=pyo.units.Mgallons) + D)*blk.unit_model.num_floc_processes

        cost_coag_inj = (E*pyo.units.convert(
            blk.unit_model.chemical_flow_mass[t0, "alum"],
            to_units=(pyo.units.lb/pyo.units.hr)) + F)*blk.unit_model.num_coag_processes

        cost_floc_inj = (G*pyo.units.convert(
            blk.unit_model.chemical_flow_mass[t0, "polymer"],
            to_units=(pyo.units.lb/pyo.units.day)) + H)*blk.unit_model.num_floc_injection_processes

        expr = (
            pyo.units.convert(
                cost_rapid_mix,
                to_units=blk.config.flowsheet_costing_block.base_currency) +
            pyo.units.convert(
                cost_floc,
                to_units=blk.config.flowsheet_costing_block.base_currency) +
            pyo.units.convert(
                cost_coag_inj,
                to_units=blk.config.flowsheet_costing_block.base_currency) +
            pyo.units.convert(
                cost_floc_inj,
                to_units=blk.config.flowsheet_costing_block.base_currency))

        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_deep_well_injection(blk):
        """
        General method for costing deep well injection processes. Capital cost
        is based on the cost of pump and pipe.
        This method also registers the electricity demand as a costed flow.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["well_pump_cost",
             "pipe_cost_basis",
             "flow_exponent"])

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        cost_well_pump = A

        cost_pipe = (B * blk.unit_model.pipe_distance[t0] *
                     blk.unit_model.pipe_diameter[t0])

        cost_total = pyo.units.convert(
            cost_well_pump + cost_pipe,
            to_units=blk.config.flowsheet_costing_block.base_currency)

        Q = pyo.units.convert(blk.unit_model.properties[t0].flow_vol,
                              to_units=pyo.units.m**3/pyo.units.hour)

        sizing_term = (Q / blk.unit_model.flow_basis[t0])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Call general power law costing method
        ZeroOrderCostingData._general_power_law_form(
            blk, cost_total, C, sizing_term, factor)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_dmbr(blk):
        """
        General method for costing dynamic membrane bioreactor. Capital cost
        is based on the cost of membrane area.

        This method also registers the electricity demand as a costed flow.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["water_flux",
             "reactor_cost"])

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        expr = pyo.units.convert(
            blk.unit_model.properties_treated[t0].flow_vol / A * B,
            to_units=blk.config.flowsheet_costing_block.base_currency)

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_fixed_bed(blk):
        """
        General method for costing fixed bed units. This primarily calls the
        cost_power_law_flow method.

        This method also registers demand for a number of additional material
        flows.
        """
        t0 = blk.flowsheet().time.first()

        ZeroOrderCostingData.cost_power_law_flow(blk)

        # Register flows - electricity already done by cost_power_law_flow
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.acetic_acid_demand[t0], "acetic_acid")
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.phosphoric_acid_demand[t0], "phosphoric_acid")
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.ferric_chloride_demand[t0], "ferric_chloride")
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.activated_carbon_demand[t0], "activated_carbon")
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.sand_demand[t0], "sand")
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.anthracite_demand[t0], "anthracite")
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.cationic_polymer_demand[t0], "cationic_polymer")

    def cost_gac(blk):
        """
        General method for costing granular activated carbon processes. Capital
        cost is based on the inlet flow rate of liquid and the empty bed
        contacting time.

        This method also registers electricity and activated carbon consumption
        as costed flows.
        """
        t0 = blk.flowsheet().time.first()

        Q = blk.unit_model.properties_in[t0].flow_vol
        T = blk.unit_model.empty_bed_contact_time

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        A, B, C = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter",
             "capital_b_parameter",
             "capital_c_parameter"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Call general power law costing method
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        expr = (pyo.units.convert(
                    A * Q,
                    to_units=blk.config.flowsheet_costing_block.base_currency) +
                pyo.units.convert(
                    B * T,
                    to_units=blk.config.flowsheet_costing_block.base_currency) +
                pyo.units.convert(
                    C * Q * T,
                    to_units=blk.config.flowsheet_costing_block.base_currency))

        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.activated_carbon_demand[t0], "activated_carbon")

    def cost_mabr(blk):
        """
        General method for costing membrane aerated biofilm reactor. Capital cost
        is based on the cost of reactor and blower.

        This method also registers the electricity demand as a costed flow.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C, D = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["specific_removal",
             "reactor_cost",
             "specific_air_flow",
             "blower_cost"])

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        DCC_reactor = pyo.units.convert(
            blk.unit_model.properties_treated[t0].flow_vol *
            blk.unit_model.properties_treated[t0].conc_mass_comp["ammonium_as_nitrogen"] / A * B,
            to_units=blk.config.flowsheet_costing_block.base_currency)

        DCC_blower = pyo.units.convert(blk.unit_model.blower_size * C * D, to_units=blk.config.flowsheet_costing_block.base_currency)

        expr = DCC_reactor + DCC_blower

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_ion_exchange(blk):
        """
        General method for costing ion exchange units. Capital cost is based on
        the both inlet flow and TDS.

        This method also registers the NaCl demand, resin replacement and
        electricity demand as costed flows.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C, D = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter",
             "capital_b_parameter",
             "capital_c_parameter",
             "capital_d_parameter"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        ln_Q = pyo.log(pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol /
            (pyo.units.m**3/pyo.units.hour),
            to_units=pyo.units.dimensionless))
        T = pyo.units.convert(
            blk.unit_model.properties_in[t0].conc_mass_comp["tds"] /
            (pyo.units.mg/pyo.units.liter),
            to_units=pyo.units.dimensionless)

        expr = pyo.units.convert(
            pyo.exp(A + B*ln_Q + C*T + D*ln_Q*T)*pyo.units.USD_2017,
            to_units=blk.config.flowsheet_costing_block.base_currency)

        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.NaCl_flowrate[t0], "sodium_chloride")
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.resin_demand[t0], "ion_exchange_resin")

    def cost_iron_and_manganese_removal(blk):
        """
        General method for costing iron and manganese removal processes. Capital cost
        is based on the cost of air blower, backwash and dual media filter.

        This method also registers the electricity demand as a costed flow.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C, D, E, F = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_blower_a_parameter",
             "capital_backwash_a_parameter",
             "capital_backwash_b_parameter",
             "capital_filter_a_parameter",
             "capital_filter_b_parameter",
             "flow_exponent"])

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        cost_blower = A

        cost_backwash = B + C*pyo.units.convert(blk.unit_model.filter_surf_area,
                                                to_units=pyo.units.ft**2)

        cost_filter = D + E * pyo.units.convert(blk.unit_model.filter_surf_area,
                                                to_units=pyo.units.ft**2)

        cost_total = pyo.units.convert(
            cost_blower + cost_backwash + cost_filter*blk.unit_model.num_filter_units,
            to_units=blk.config.flowsheet_costing_block.base_currency)

        Q = pyo.units.convert(blk.unit_model.properties_in[t0].flow_vol,
                              to_units=pyo.units.m**3/pyo.units.hour)

        sizing_term = (Q / blk.unit_model.flow_basis[t0])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Call general power law costing method
        ZeroOrderCostingData._general_power_law_form(
            blk, cost_total, F, sizing_term, factor)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_sedimentation(blk):
        """
        General method for costing sedimentaion processes. Capital cost is
        based on the surface area of the basin.
        """
        t0 = blk.flowsheet().time.first()
        sizing_term = (blk.unit_model.basin_surface_area[t0] /
                       pyo.units.foot**2)

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        A, B = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter", "capital_b_parameter"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Call general power law costing method
        ZeroOrderCostingData._general_power_law_form(
            blk, A, B, sizing_term, factor)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_storage_tank(blk):
        """
        General method for costing storage tanks. Capital cost is based on the
        volume of the tank.
        """
        t0 = blk.flowsheet().time.first()
        sizing_term = (blk.unit_model.tank_volume[t0] / pyo.units.m**3)

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter", "capital_b_parameter"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Call general power law costing method
        ZeroOrderCostingData._general_power_law_form(
            blk, A, B, sizing_term, factor)

    def cost_surface_discharge(blk):
        """
        General method for costing surface discharge. Capital cost is based on
        construction and pipe costs.
        """

        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, pipe_cost_basis, ref_state = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter", "capital_b_parameter", "pipe_cost_basis", "reference_state"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        expr = (
            pyo.units.convert(
                A * pyo.units.convert(blk.unit_model.properties[t0].flow_vol / ref_state, to_units=pyo.units.dimensionless) ** B,
                to_units=blk.config.flowsheet_costing_block.base_currency) +
            pyo.units.convert(
                pipe_cost_basis * blk.unit_model.pipe_distance[t0] * blk.unit_model.pipe_diameter[t0],
                to_units=blk.config.flowsheet_costing_block.base_currency))

        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_ozonation(blk):
        """
        General method for costing ozone addition. Capital cost is
        based on the inlet flowrate and dosage of ozone.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C, D = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["ozone_capital_a_parameter",
             "ozone_capital_b_parameter",
             "ozone_capital_c_parameter",
             "ozone_capital_d_parameter"])
        # Get costing term for ozone addition
        expr = ZeroOrderCostingData._get_ozone_capital_cost(blk, A, B, C, D)
        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]
        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        # Add cost variable
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_ozonation_aop(blk):
        """
        General method for costing ozonation with AOP. Capital cost is
        based on the inlet flowrate, dosage of ozone and flow rate of H2O2.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C, D, E, F = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["ozone_capital_a_parameter",
             "ozone_capital_b_parameter",
             "ozone_capital_c_parameter",
             "ozone_capital_d_parameter",
             "aop_capital_a_parameter",
             "aop_capital_b_parameter"])

        # Add cost variable
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        # Get costing term for ozone addition
        expr = ZeroOrderCostingData._get_ozone_capital_cost(blk, A, B, C, D)

        # Add costing term for AOP addition
        expr += ZeroOrderCostingData._get_aop_capital_cost(blk, E, F)

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.chemical_flow_mass[t0], "hydrogen_peroxide")

    def cost_uv(blk):
        """
        General method for costing UV reactor units. Capital cost is based on
        the inlet flow, UV reduced equivalent dosage, and UV transmittance at
        the inlet.
        """
        t0 = blk.flowsheet().time.first()
        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C, D = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["uv_capital_a_parameter",
             "uv_capital_b_parameter",
             "uv_capital_c_parameter",
             "uv_capital_d_parameter"])

        expr = ZeroOrderCostingData._get_uv_capital_cost(blk, A, B, C, D)

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_uv_aop(blk):
        t0 = blk.flowsheet().time.first()

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")
        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C, D, E, F = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["uv_capital_a_parameter",
             "uv_capital_b_parameter",
             "uv_capital_c_parameter",
             "uv_capital_d_parameter",
             "aop_capital_a_parameter",
             "aop_capital_b_parameter"])

        expr = ZeroOrderCostingData._get_uv_capital_cost(blk, A, B, C, D)
        expr += ZeroOrderCostingData._get_aop_capital_cost(blk, E, F)

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]
        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

        #TODO: Check whether chemical flow cost was accounted for originally
        # and if should be in case study verification
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.chemical_flow_mass[t0], "hydrogen_peroxide")

    def cost_evaporation_pond(blk):
        """
        General method for costing evaporation pond. Capital cost is based on the pond area and
        other pond construction parameters.
        """

        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, C, D, E, liner_thickness, land_cost, land_clearing_cost = \
            _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["cost_per_acre_a_parameter", "cost_per_acre_b_parameter", 
            "cost_per_acre_c_parameter", "cost_per_acre_d_parameter", 
            "cost_per_acre_e_parameter", "liner_thickness", "land_cost", 
            "land_clearing_cost"])

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        expr = pyo.units.convert(
            blk.unit_model.adj_area[t0]*
            (A + B*liner_thickness +
            C*land_cost+
            D*land_clearing_cost+
            E*blk.unit_model.dike_height[t0]), 
            to_units=blk.config.flowsheet_costing_block.base_currency)

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

    
    def cost_filter_press(blk):
        """
        General method for costing belt filter press. Capital cost is a function
        of flow in gal/hr.
        """
        t0 = blk.flowsheet().time.first()
        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        Q = pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol,
            to_units=pyo.units.gal/pyo.units.hr)

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter",
             "capital_b_parameter"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        expr = pyo.units.convert(
            A*Q + B,
            to_units=blk.config.flowsheet_costing_block.base_currency)

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")
            

    def cost_landfill(blk):
        """
        General method for costing landfill. Capital cost is based on the total mass and
        capacity basis.
        """

        t0 = blk.flowsheet().time.first()
        sizing_term = (blk.unit_model.total_mass[t0] / blk.unit_model.capacity_basis[t0])

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter", "capital_b_parameter"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Call general power law costing method
        ZeroOrderCostingData._general_power_law_form(
            blk, A, B, sizing_term, factor)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def cost_well_field(blk):
        """
        General method for costing well fields. Capital cost is based on well field
        construction and pipe costs.
        """

        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = \
            blk.unit_model.config.database.get_unit_operation_parameters(
                blk.unit_model._tech_type,
                subtype=blk.unit_model.config.process_subtype)

        # Get costing parameter sub-block for this technology
        A, B, pipe_cost_basis = _get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter", "capital_b_parameter", "pipe_cost_basis"])

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        Q = pyo.units.convert(
            blk.unit_model.properties[t0].flow_vol /
            (pyo.units.m**3/pyo.units.hour),
            to_units=pyo.units.dimensionless)
        expr = pyo.units.convert(
            A*Q**B + (pipe_cost_basis *
            blk.unit_model.pipe_distance[t0] * blk.unit_model.pipe_diameter[t0]),
            to_units=blk.config.flowsheet_costing_block.base_currency)

        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity")

    def _get_ozone_capital_cost(blk, A, B, C, D):
        """
        Generate expressions for capital cost of ozonation system.
        """
        t0 = blk.flowsheet().time.first()

        ln_Q = pyo.log(pyo.units.convert(
                blk.unit_model.properties_in[t0].flow_vol /
                (pyo.units.m**3/pyo.units.hour),
                to_units=pyo.units.dimensionless))
        dosage = pyo.units.convert(
            blk.unit_model.ozone_consumption[t0] /
            (pyo.units.mg/pyo.units.liter),
            to_units=pyo.units.dimensionless)

        expr = pyo.units.convert(
            A + B*dosage + C*ln_Q + D*dosage*ln_Q,
            to_units=blk.config.flowsheet_costing_block.base_currency)

        return expr

    def _get_uv_capital_cost(blk, A, B, C, D):
        """
        Generate expression for capital cost of UV reactor.
        """
        t0 = blk.flowsheet().time.first()

        Q = pyo.units.convert(
                pyo.units.convert(blk.unit_model.properties_in[t0].flow_vol,
                                  to_units=pyo.units.Mgallons/pyo.units.day)
                / (pyo.units.Mgallons/pyo.units.day),
                to_units=pyo.units.dimensionless)

        uv_dose = pyo.units.convert(
            blk.unit_model.uv_reduced_equivalent_dose[t0] /
            (pyo.units.mJ / pyo.units.cm ** 2),
            to_units=pyo.units.dimensionless)

        uvt_in = blk.unit_model.uv_transmittance_in[t0]

        expr = pyo.units.convert(
            A * Q + B * uv_dose * Q + C * (Q * uvt_in) ** 7 +
            D * uv_dose * Q * uvt_in,
            to_units=blk.config.flowsheet_costing_block.base_currency)

        return expr

    def _get_aop_capital_cost(blk, A, B):
        """
        Generate expression for capital cost due to AOP addition.
        """
        t0 = blk.flowsheet().time.first()

        chemical_flow_mass = pyo.units.convert(
            blk.unit_model.chemical_flow_mass[t0],
            to_units=pyo.units.lb/pyo.units.day)
        expr = pyo.units.convert(
            A*pyo.units.convert(chemical_flow_mass /
                                (pyo.units.lb/pyo.units.day),
                                to_units=pyo.units.dimensionless) ** B,
            to_units=blk.config.flowsheet_costing_block.base_currency)

        return expr

    def _general_power_law_form(blk, A, B, sizing_term, factor=None):
        """
        General method for building power law costing expressions.
        """
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation")

        expr = pyo.units.convert(
            A * pyo.units.convert(sizing_term,
                                  to_units=pyo.units.dimensionless) ** B,
            to_units=blk.config.flowsheet_costing_block.base_currency)

        if factor == "TPEC":
            expr *= blk.config.flowsheet_costing_block.TPEC
        elif factor == "TIC":
            expr *= blk.config.flowsheet_costing_block.TIC

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == expr)
    # -------------------------------------------------------------------------
    # Map costing methods to unit model classes
    unit_mapping = {ZeroOrderBase: cost_power_law_flow,
                    BrineConcentratorZO: cost_brine_concentrator,
                    ChemicalAdditionZO: cost_chemical_addition,
                    ChlorinationZO: cost_chlorination,
                    CoagulationFlocculationZO: cost_coag_and_floc,
                    DeepWellInjectionZO: cost_deep_well_injection,
                    DMBRZO: cost_dmbr,
                    FixedBedZO: cost_fixed_bed,
                    GACZO: cost_gac,
                    LandfillZO: cost_landfill,
                    MABRZO: cost_mabr,
                    IonExchangeZO: cost_ion_exchange,
                    IronManganeseRemovalZO: cost_iron_and_manganese_removal,
                    OzoneZO: cost_ozonation,
                    OzoneAOPZO: cost_ozonation_aop,
                    SedimentationZO: cost_sedimentation,
                    StorageTankZO: cost_storage_tank,
                    SurfaceDischargeZO: cost_surface_discharge,
                    UVZO: cost_uv,
                    UVAOPZO: cost_uv_aop,
                    EvaporationPondZO: cost_evaporation_pond,
                    FilterPressZO: cost_filter_press,
                    WellFieldZO: cost_well_field,
                    }


def _get_tech_parameters(blk, parameter_dict, subtype, param_list):
    """
    First, need to check to see if a Block with parameters for this technology
    exists.
    Second, to handle technology subtypes all parameters need to be indexed by
    subtype. We will dynamically add subtypes to the indexing set and Vars as
    required.
    """
    # Check to see in parameter Block already exists
    try:
        # Try to get parameter Block from costing package
        pblock = getattr(blk.config.flowsheet_costing_block,
                         blk.unit_model._tech_type)
    except AttributeError:
        # Parameter Block for this technology haven't been added yet, create
        pblock = pyo.Block()

        # Add block to FlowsheetCostingBlock
        blk.config.flowsheet_costing_block.add_component(
            blk.unit_model._tech_type, pblock)

        # Add subtype Set to Block
        pblock.subtype_set = pyo.Set()

        # Add required Vars
        for p in param_list:
            try:
                vobj = pyo.Var(
                    pblock.subtype_set,
                    units=getattr(
                        pyo.units,
                        parameter_dict[
                            "capital_cost"][p]["units"]))
                pblock.add_component(p, vobj)
            except KeyError:
                raise KeyError(
                    "Error when trying to retrieve costing parameter"
                    " from {p} database. Please check the YAML "
                    "file for this technology for errors.")

    # Check to see if required subtype is in subtype_set
    vlist = []
    if subtype not in pblock.subtype_set:
        # Need to add subtype and set Vars
        pblock.subtype_set.add(subtype)

        # Set vars
        for p in param_list:
            vobj = getattr(pblock, p)
            vobj[subtype].fix(
                float(parameter_dict[
                    "capital_cost"][p]["value"]) *
                getattr(pyo.units,
                        parameter_dict[
                            "capital_cost"][p]["units"]))
            vlist.append(vobj[subtype])
    else:
        for p in param_list:
            vobj = getattr(pblock, p)
            vlist.append(vobj[subtype])

    return tuple(x for x in vlist)


def _load_case_study_definition(self):
    """
    Load data from case study definition file into a Python dict.

    If users did not provide a definition file as a config argument, the
    default definition from the WaterTap techno-economic database is used.
    """
    source_file = self.config.case_study_definition
    if source_file is None:
        source_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "..", "data", "techno_economic", "default_case_study.yaml")

    try:
        with open(source_file, "r") as f:
            lines = f.read()
            f.close()
    except OSError:
        raise OSError(
            "Could not find specified case study definition file. "
            "Please check the path provided.")

    return yaml.load(lines, yaml.Loader)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for general zero-order property package
"""
import pytest
from io import StringIO
import os

from idaes.core import declare_process_block_class, FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core.util import get_solver
import idaes.core.util.scaling as iscale
from pyomo.environ import (check_optimal_termination,
                           ConcreteModel,
                           Constraint,
                           Param,
                           Set,
                           value,
                           Var)
from pyomo.network import Port
from pyomo.util.check_units import assert_units_consistent


from watertap.core import (
    Database, WaterParameterBlock, WaterStateBlock, ZeroOrderBaseData)
from watertap.core.zero_order_sido_reactive import (
    build_sido_reactive,
    initialize_sidor,
    calculate_scaling_factors_sidor,
    _get_Q_sidor)

solver = get_solver()

local_path = os.path.dirname(os.path.abspath(__file__))


@declare_process_block_class("DerivedSIDOR")
class DerivedSIDORData(ZeroOrderBaseData):
    def build(self):
        super().build()

        self._tech_type = "test_sidor_data"

        build_sido_reactive(self)


class TestSIDOR:
    @pytest.fixture(scope="module")
    def model(self):
        m = ConcreteModel()

        m.db = Database(dbpath=local_path)

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        m.fs.unit = DerivedSIDOR(
            default={"property_package": m.fs.water_props,
                     "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1000)
        m.fs.unit.inlet.flow_mass_comp[0, "A"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "B"].fix(20)
        m.fs.unit.inlet.flow_mass_comp[0, "C"].fix(30)

        m.fs.unit.load_parameters_from_database(use_default_removal=True)

        return m

    @pytest.mark.unit
    def test_private_attributes(self, model):
        assert model.fs.unit._has_recovery_removal is True
        assert model.fs.unit._fixed_perf_vars == []
        assert model.fs.unit._initialize is initialize_sidor
        assert model.fs.unit._scaling is calculate_scaling_factors_sidor
        assert model.fs.unit._get_Q is _get_Q_sidor
        assert model.fs.unit._stream_table_dict == {
            "Inlet": model.fs.unit.inlet,
            "Treated": model.fs.unit.treated,
            "Byproduct": model.fs.unit.byproduct}
        assert model.fs.unit._perf_var_dict == {
            "Water Recovery": model.fs.unit.recovery_frac_mass_H2O,
            "Solute Removal": model.fs.unit.removal_frac_mass_solute,
            "Reaction Extent": model.fs.unit.extent_of_reaction}

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.fs.unit.properties_in, WaterStateBlock)
        assert isinstance(model.fs.unit.properties_treated, WaterStateBlock)
        assert isinstance(model.fs.unit.properties_byproduct, WaterStateBlock)

        assert isinstance(model.fs.unit.inlet, Port)
        assert isinstance(model.fs.unit.treated, Port)
        assert isinstance(model.fs.unit.byproduct, Port)

        assert isinstance(model.fs.unit.recovery_frac_mass_H2O, Var)
        assert len(model.fs.unit.recovery_frac_mass_H2O) == 1
        assert isinstance(model.fs.unit.removal_frac_mass_solute, Var)
        assert len(model.fs.unit.removal_frac_mass_solute) == 3

        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert len(model.fs.unit.water_recovery_equation) == 1
        assert isinstance(model.fs.unit.water_balance, Constraint)
        assert len(model.fs.unit.water_balance) == 1
        assert isinstance(model.fs.unit.solute_removal_equation, Constraint)
        assert len(model.fs.unit.solute_removal_equation) == 3
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)
        assert len(model.fs.unit.solute_treated_equation) == 3

        assert isinstance(model.fs.unit.reaction_set, Set)
        assert isinstance(model.fs.unit.generation_ratio, Param)

        assert isinstance(model.fs.unit.reaction_conversion, Var)
        assert isinstance(model.fs.unit.extent_of_reaction, Var)

        assert isinstance(model.fs.unit.reaction_extent_equation, Constraint)

        for r in model.fs.unit.reaction_set:
            assert r in ["Rxn1", "Rxn2"]
            assert (0, r) in model.fs.unit.reaction_conversion
            assert (0, r) in model.fs.unit.extent_of_reaction
            assert (0, r) in model.fs.unit.reaction_extent_equation
            for j in model.fs.water_props.component_list:
                assert (r, j) in model.fs.unit.generation_ratio

    @pytest.mark.unit
    def test_key_components(self, model):
        assert (str(model.fs.unit.properties_in[0].flow_mass_comp["A"]) in
                str(model.fs.unit.reaction_extent_equation[0, "Rxn1"].body))
        assert (str(model.fs.unit.properties_in[0].flow_mass_comp["B"]) in
                str(model.fs.unit.reaction_extent_equation[0, "Rxn2"].body))

    @pytest.mark.unit
    def test_loading_data(self, model):
        # RXn 1 is all conversion ratios
        # Rxn 2 is all stocihiometry (incl. water)
        cfactor = {"Rxn1": {"A": -1, "B": 1, "C": 0, "H2O": 0},
                   "Rxn2": {"A": 0, "B": -1, "C": 22*2/20, "H2O": -1*18/20}}

        for (r, j), p in model.fs.unit.generation_ratio.items():
            assert value(p) == cfactor[r][j]

        assert model.fs.unit.reaction_conversion[0, "Rxn1"].value == 0.8
        assert model.fs.unit.reaction_conversion[0, "Rxn2"].value == 0.1

        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 0.85
        assert model.fs.unit.removal_frac_mass_solute[0, "A"].value == 0.5
        assert model.fs.unit.removal_frac_mass_solute[0, "B"].value == 0.4
        assert model.fs.unit.removal_frac_mass_solute[0, "C"].value == 0.0

    @pytest.mark.unit
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model)

    @pytest.mark.component
    def test_scaling(self, model):
        iscale.calculate_scaling_factors(model)

        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.water_recovery_equation[0]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.water_balance[0]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.solute_removal_equation[0, "A"]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.solute_removal_equation[0, "B"]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.solute_removal_equation[0, "C"]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.solute_treated_equation[0, "A"]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.solute_treated_equation[0, "B"]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.solute_treated_equation[0, "C"]) == 1e5

        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.reaction_extent_equation[0, "Rxn1"]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.reaction_extent_equation[0, "Rxn2"]) == 1e5

    @pytest.mark.component
    def test_initialization(self, model):
        initialization_tester(model)

    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, model):
        model.fs.unit.treated.display()
        model.fs.unit.byproduct.display()
        assert (pytest.approx((1000-0.1*20*18/20)*0.85, rel=1e-5) ==
                value(model.fs.unit.treated.flow_mass_comp[0, "H2O"]))
        assert (pytest.approx((1000-0.1*20*18/20)*0.15, rel=1e-5) ==
                value(model.fs.unit.byproduct.flow_mass_comp[0, "H2O"]))

        assert (pytest.approx((10*0.2)*(1-0.5), rel=1e-5) ==
                value(model.fs.unit.treated.flow_mass_comp[0, "A"]))
        assert (pytest.approx((10*0.2)*0.5, rel=1e-5) ==
                value(model.fs.unit.byproduct.flow_mass_comp[0, "A"]))
        assert (pytest.approx((20+0.8*10-0.1*20)*(1-0.4), rel=1e-5) ==
                value(model.fs.unit.treated.flow_mass_comp[0, "B"]))
        assert (pytest.approx((20+0.8*10-0.1*20)*0.4, rel=1e-5) ==
                value(model.fs.unit.byproduct.flow_mass_comp[0, "B"]))
        assert (pytest.approx(30+0.1*20*2*22/20, rel=1e-5) ==
                value(model.fs.unit.treated.flow_mass_comp[0, "C"]))
        assert (pytest.approx(0, rel=1e-5) ==
                value(model.fs.unit.byproduct.flow_mass_comp[0, "C"]))

    # Conservation would be similar to above, as need to calculate generation

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()
        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                    : Value   : Fixed : Bounds
    Reaction Extent [Rxn1] :  8.0000 : False : (None, None)
    Reaction Extent [Rxn2] :  2.0000 : False : (None, None)
        Solute Removal [A] : 0.50000 :  True : (0, None)
        Solute Removal [B] : 0.40000 :  True : (0, None)
        Solute Removal [C] :  0.0000 :  True : (0, None)
            Water Recovery : 0.85000 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                            Inlet  Treated  Byproduct
    Volumetric Flowrate    1.0600 0.89947     0.16113
    Mass Concentration H2O 943.40  943.30      929.25
    Mass Concentration A   9.4340  1.1118      6.2062
    Mass Concentration B   18.868  17.344      64.544
    Mass Concentration C   28.302  38.245  6.2062e-13
====================================================================================
"""

        assert output == stream.getvalue()


class TestSIDORErrors:
    @pytest.mark.unit
    def test_reaction_list(self):
        m = ConcreteModel()

        m.db = Database(dbpath=local_path)

        # Load data from YAML so that we can modify it
        m.db._get_technology("test_sidor_data")
        m.db._cached_files["test_sidor_data"]["default"] = {}

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        with pytest.raises(
                KeyError,
                match="fs.unit - database provided does not contain a list of "
                "reactions for this technology."):
            m.fs.unit = DerivedSIDOR(
                default={"property_package": m.fs.water_props,
                         "database": m.db})

    @pytest.mark.unit
    def test_missing_conversion(self):
        m = ConcreteModel()

        m.db = Database(dbpath=local_path)

        # Load data from YAML so that we can modify it
        m.db._get_technology("test_sidor_data")
        m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"] = {}

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        with pytest.raises(
                KeyError,
                match="fs.unit - database provided does not "
                "contain an entry for conversion for reaction Rxn3."):
            m.fs.unit = DerivedSIDOR(
                default={"property_package": m.fs.water_props,
                         "database": m.db})

    @pytest.mark.unit
    def test_missing_key_reactant(self):
        m = ConcreteModel()

        m.db = Database(dbpath=local_path)

        # Load data from YAML so that we can modify it
        m.db._get_technology("test_sidor_data")
        m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"] = {}
        R3 = m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"]
        R3["conversion"] = 0.5

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        with pytest.raises(
                KeyError,
                match="fs.unit - database provided does not "
                "contain an entry for key_reactant for reaction Rxn3."):
            m.fs.unit = DerivedSIDOR(
                default={"property_package": m.fs.water_props,
                         "database": m.db})

    @pytest.mark.unit
    def test_invlaid_key_reactant(self):
        m = ConcreteModel()

        m.db = Database(dbpath=local_path)

        # Load data from YAML so that we can modify it
        m.db._get_technology("test_sidor_data")
        m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"] = {}
        R3 = m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"]
        R3["conversion"] = 0.5
        R3["key_reactant"] = "foo"

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        with pytest.raises(
                ValueError,
                match="fs.unit - key_reactant foo for reaction Rxn3 "
                "is not in the component list used by the assigned property "
                "package."):
            m.fs.unit = DerivedSIDOR(
                default={"property_package": m.fs.water_props,
                         "database": m.db})

    @pytest.mark.unit
    def test_missing_stoichiometry(self):
        m = ConcreteModel()

        m.db = Database(dbpath=local_path)

        # Load data from YAML so that we can modify it
        m.db._get_technology("test_sidor_data")
        m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"] = {}
        R3 = m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"]
        R3["conversion"] = 0.5
        R3["key_reactant"] = "A"

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        with pytest.raises(
                KeyError,
                match="fs.unit - database provided does not "
                "contain an entry for stoichiometry for reaction Rxn3."):
            m.fs.unit = DerivedSIDOR(
                default={"property_package": m.fs.water_props,
                         "database": m.db})

    @pytest.mark.unit
    def test_ratio_and_order(self):
        m = ConcreteModel()

        m.db = Database(dbpath=local_path)

        # Load data from YAML so that we can modify it
        m.db._get_technology("test_sidor_data")
        m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"] = {}
        R3 = m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"]
        R3["conversion"] = 0.5
        R3["key_reactant"] = "A"
        R3["stoichiometry"] = {"A": {"order": 1, "conversion_ratio": 1}}

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        with pytest.raises(
                RuntimeError,
                match="fs.unit - database provides entries for both "
                "conversion_ratio and reaction order in reaction Rxn3. "
                "Please provide only one or the other."):
            m.fs.unit = DerivedSIDOR(
                default={"property_package": m.fs.water_props,
                         "database": m.db})

    @pytest.mark.unit
    def test_no_ratio_or_order(self):
        m = ConcreteModel()

        m.db = Database(dbpath=local_path)

        # Load data from YAML so that we can modify it
        m.db._get_technology("test_sidor_data")
        m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"] = {}
        R3 = m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"]
        R3["conversion"] = 0.5
        R3["key_reactant"] = "A"
        R3["stoichiometry"] = {"A": {}}

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        with pytest.raises(
                RuntimeError,
                match="fs.unit - database provided does not "
                "contain any information for conversion_ratio or reaction "
                "order w.r.t. species A in reaction Rxn3."):
            m.fs.unit = DerivedSIDOR(
                default={"property_package": m.fs.water_props,
                         "database": m.db})

    @pytest.mark.unit
    def test_order_no_mw(self):
        m = ConcreteModel()

        m.db = Database(dbpath=local_path)

        # Load data from YAML so that we can modify it
        m.db._get_technology("test_sidor_data")
        m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"] = {}
        R3 = m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"]
        R3["conversion"] = 0.5
        R3["key_reactant"] = "B"
        R3["stoichiometry"] = {"A": {"order": 1},
                               "B": {"order": 1, "molecular_weight": 1}}

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        with pytest.raises(
                KeyError,
                match="fs.unit - database provided does not "
                "contain an entry for molecular_weight w.r.t. "
                "species A in reaction Rxn3."):
            m.fs.unit = DerivedSIDOR(
                default={"property_package": m.fs.water_props,
                         "database": m.db})

    @pytest.mark.unit
    def test_order_no_key_order(self):
        m = ConcreteModel()

        m.db = Database(dbpath=local_path)

        # Load data from YAML so that we can modify it
        m.db._get_technology("test_sidor_data")
        m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"] = {}
        R3 = m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"]
        R3["conversion"] = 0.5
        R3["key_reactant"] = "B"
        R3["stoichiometry"] = {"A": {"order": 1, "molecular_weight": 1}}

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        with pytest.raises(
                KeyError,
                match="fs.unit - database provided does not "
                "contain an entry for order w.r.t. species "
                "B in reaction Rxn3."):
            m.fs.unit = DerivedSIDOR(
                default={"property_package": m.fs.water_props,
                         "database": m.db})

    @pytest.mark.unit
    def test_order_no_key_mw(self):
        m = ConcreteModel()

        m.db = Database(dbpath=local_path)

        # Load data from YAML so that we can modify it
        m.db._get_technology("test_sidor_data")
        m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"] = {}
        R3 = m.db._cached_files[
            "test_sidor_data"]["default"]["reactions"]["Rxn3"]
        R3["conversion"] = 0.5
        R3["key_reactant"] = "B"
        R3["stoichiometry"] = {"A": {"order": 1, "molecular_weight": 1},
                               "B": {"order": 1}}

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        with pytest.raises(
                KeyError,
                match="fs.unit - database provided does not "
                "contain an entry for molecular_weight w.r.t. "
                "species B in reaction Rxn3."):
            m.fs.unit = DerivedSIDOR(
                default={"property_package": m.fs.water_props,
                         "database": m.db})



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of an electrodialysis reversal unit.
"""

from pyomo.environ import Constraint, units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_sido, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("ElectrodialysisReversalZO")
class ElectrodialysisReversalZOData(ZeroOrderBaseData):
    """
    Zero-Order model for an electrodialysis reversal unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "electrodialysis_reversal"

        build_sido(self)

        if "tds" not in self.config.property_package.solute_set:
            raise KeyError(
                'TDS must be included in the solute list for determining'
                ' electricity intensity and power consumption of the electrodialysis '
                'reversal unit.')

        self.elec_coeff_1 = Var(units=pyunits.kWh/pyunits.m**3,
                                  doc="Constant 1 in electricity intensity equation")
        self.elec_coeff_2 = Var(units=pyunits.L/pyunits.mg*pyunits.kWh/pyunits.m**3,
                                  doc="Constant 2 in electricity intensity equation")

        self._fixed_perf_vars.append(self.elec_coeff_1)
        self._fixed_perf_vars.append(self.elec_coeff_2)

        self.electricity = Var(
            self.flowsheet().config.time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Power consumption of brine concentrator")
        self.electricity_intensity = Var(
            self.flowsheet().config.time,
            units=pyunits.kWh/pyunits.m**3,
            doc="Specific energy consumption with respect to feed flowrate")

        @self.Constraint(self.flowsheet().config.time,
                         doc="Electricity intensity constraint")
        def electricity_intensity_constraint(b, t):
            tds_in = pyunits.convert(b.properties_in[t].conc_mass_comp["tds"],
                                     to_units=pyunits.mg/pyunits.L)
            return (b.electricity_intensity[t] ==
                    b.elec_coeff_1
                    + b.elec_coeff_2 * tds_in)

        @self.Constraint(self.flowsheet().config.time,
                         doc="Power consumption constraint")
        def electricity_constraint(b, t):
            q_in = pyunits.convert(b.properties_in[t].flow_vol,
                                   to_units=pyunits.m**3/pyunits.hour)
            return (b.electricity[t] ==
                    b.electricity_intensity[t] * q_in)

        self._perf_var_dict["Power Consumption (kW)"] = self.electricity
        self._perf_var_dict[
            "Electricity intensity per Inlet Flowrate  (kWh/m3)"] = \
            self.electricity_intensity
